# This is a list of implicit operands which are read/written by
# various x86 instructions. Note that modifications to the stack
# register are mentioned here, but that additional information on
# the effect an instruction has on the stack is contained in the
# x86_insn_t 'stack_mod' and 'stack_mod_val' fields. Use of the
# eflags register, i.e. setting, clearing, and testing flags, is
# not recorded here but rather in the flags_set and flags_tested
# fields of the x86_insn_t.

# Conventions: Register operands which are aliases of another register 
# operand (e.g. AX in one operand and AL in another) assume that the
# operands are different registers and that alias tracking will resolve
# data flow. This means that something like
#   	mov ax, al
# would have 'write only' access for AX and 'read only' access for AL,
# even though both AL and AX are read and written 

# TODO: add a datatype and operand size field

implicit_operands = (
	None,							# invalid
	# 37 : AAA : rw AL 
	# 3F : AAS : rw AL 
	( { 'access':'rw-', 'register':REG_AL_INDEX },	), 	# aaa 

	# D5 0A, D5 (ib) : AAD : rw AX 
	# D4 0A, D4 (ib) : AAM : rw AX 
	( { 'access':'rw-', 'register':REG_AX_INDEX }, ),	# aad 

	# E8, FF, 9A, FF : CALL : rw ESP, rw EIP 
	# C2, C3, CA, CB : RET  : rw ESP, rw EIP
	( { 'access':'rw-', 'register':REG_EIP_INDEX }, 
	  { 'access':'rw-', 'register':REG_ESP_INDEX } ),	# call 

	# 98 : CBW : r AL, rw AX 
	( { 'access':'rw-', 'register':REG_AX_INDEX },
	  { 'access':'r--', 'register':REG_AL_INDEX} ),	# cbw 

	# 98 : CWDE : r AX, rw EAX 
	( { 'access':'rw-', 'register':REG_EAX_INDEX },
	  { 'access':'r--', 'register':REG_AX_INDEX } ),	# cwde 

	# 0F 06 : CLTS : rw CR0 
	( { 'access':'rw-', 'register':REG_CR0_INDEX}, ),	# clts 

	# 0F B0 : CMPXCHG : rw AL 
	( { 'access':'rw-', 'register':REG_AL_INDEX }, ),	# cmpxchg 

	# 0F B1 : CMPXCHG : rw EAX 
	( { 'access':'rw-', 'register':REG_EAX_INDEX }, ),	# cmpxchg 

	# 0F C7 : CMPXCHG8B : rw EDX, rw EAX, r ECX, r EBX 
	( { 'access':'rw-', 'register':REG_EAX_INDEX }, 
	  { 'access':'rw-', 'register':REG_EDX_INDEX }, 
	  { 'access':'r--', 'register':REG_ECX_INDEX }, 
	  { 'access':'r--', 'register':REG_EBX_INDEX } ),	# cmpxchg8b 

	# 0F A2 : CPUID : rw EAX, w EBX, w ECX, w EDX 
	( { 'access':'rw-', 'register':REG_EAX_INDEX }, 
	  { 'access':'-w-', 'register':REG_ECX_INDEX }, 
	  { 'access':'-w-', 'register':REG_EDX_INDEX }, 
	  { 'access':'-w-', 'register':REG_EBX_INDEX } ),	# cpuid 

	# 99 : CWD/CWQ : rw EAX, w EDX 
	( { 'access':'rw-', 'register':REG_EAX_INDEX }, 
	  { 'access':'-w-', 'register':REG_EDX_INDEX } ),	# cwd 

	# 27 : DAA : rw AL 
	# 2F : DAS : rw AL 
	( { 'access':'rw-', 'register':REG_AL_INDEX }, ),	# daa 

	# F6 : DIV, IDIV : r AX, w AL, w AH 
	# FIXED: first was EAX not AX. TODO: verify.
	( { 'access':'r--', 'register':REG_AX_INDEX }, 
	   { 'access':'-w-', 'register':REG_AL_INDEX },
	   { 'access':'-w-', 'register':REG_AH_INDEX } ),	# div 

	# F7 : DIV, IDIV : rw EDX, rw EAX 
	( { 'access':'rw-', 'register':REG_EDX_INDEX }, 
	   { 'access':'rw-', 'register':REG_EAX_INDEX } ),	# div 

	# C8 : ENTER : rw ESP w EBP 
	( { 'access':'rw-', 'register':REG_ESP_INDEX }, 
	  { 'access':'r--', 'register':REG_EBP_INDEX } ),	# enter 

	# D9 F0 : F2XM1 : rw ST(0) 
	# D9 E1 : FABS : rw ST(0) 
	# D9 E0 : FCHS : rw ST(0) 
	# D9 FF : FCOS : rw ST(0)
	# D8, DA : FDIV : rw ST(0) 
	# D8, DA : FDIVR : rw ST(0) 
	# D9 F2 : FPTAN : rw ST(0) 
	# D9 FC : FRNDINT : rw ST(0) 
	# D9 FB : FSINCOS : rw ST(0) 
	# D9 FE : FSIN : rw ST(0) 
	# D9 FA : FSQRT : rw ST(0) 
	# D9 F4 : FXTRACT : rw ST(0) 
	( { 'access':'rw-', 'register':REG_ST0_INDEX }, ),	# f2xm1 

	# D8, DC, DE D9 : FCOM : r ST(0) 
	# DE, DA : FICOM : r ST(0) 
	# DF, D8 : FIST : r ST(0) 
	# D9 E4 : FTST : r ST(0) 
	# D9 E5 : FXAM : r ST(0) 
	( { 'access':'r--', 'register':REG_ST0_INDEX }, ),	# fcom 

	# D9 F3 : FPATAN : r ST(0), rw ST(1) 
	( { 'access':'r--', 'register':REG_ST0_INDEX }, ),	# fpatan 

	# D9 F8, D9 F5 : FPREM : rw ST(0) r ST(1) 
	# D9 FD : FSCALE : rw ST(0), r ST(1) 
	( { 'access':'rw-', 'register':REG_ST0_INDEX }, 	
	  { 'access':'r--', 'register':REG_ST1_INDEX } ),	# fprem 

	# DE C1 : FADDP : r ST(0), rw ST(1) 
	# DE E9 : FSUBP : r ST(0), rw ST(1) 
	# D9 F1 : FYL2X : r ST(0), rw ST(1) 
	# D9 F9 : FYL2XP1 : r ST(0), rw ST(1) 
	( { 'access':'r--', 'register':REG_ST0_INDEX },
	  { 'access':'rw-', 'register':REG_ST1_INDEX } ),	# faddp 

	# DA E9 : FUCOMPP : r ST(0), r ST(1) 
	( { 'access':'r--', 'register':REG_ST0_INDEX },
	  { 'access':'r--', 'register':REG_ST1_INDEX } ),	# fucompp 

	# F6 : IMUL : r AL, w AX 
	# F6 : MUL : r AL, w AX 
	( { 'access':'r--', 'register':REG_AL_INDEX },
	  { 'access':'-w-', 'register':REG_AX_INDEX }	 ),	# imul 

	# F7 : IMUL : rw EAX, w EDX 
	# F7 : MUL : rw EAX, w EDX 
	( { 'access':'rw-', 'register':REG_EAX_INDEX },
	  { 'access':'-w-', 'register':REG_EDX_INDEX } ),	# imul 

	# 9F : LAHF : r EFLAGS, w AH 
	( { 'access':'r--', 'register':REG_FLAGS_INDEX },
	  { 'access':'-w-', 'register':REG_AH_INDEX } ),	# lahf 

	# 0F AE : LDMXCSR : w MXCSR SSE Control Status Reg 
	( { 'access':'-w-', 'register':REG_MXCSG_INDEX }, ),	# ldmxcsr 

	# C9 : LEAVE :  rw ESP, w EBP 
	( { 'access':'rw-', 'register':REG_ESP_INDEX },
	  { 'access':'-w-', 'register':REG_EBP_INDEX } ),	# leave 

	# 0F 01 : LGDT : w GDTR 
	( { 'access':'-w-', 'register':REG_GDTR_INDEX }, ),	# lgdt 

	# 0F 01 : LIDT : w IDTR 
	( { 'access':'-w-', 'register':REG_IDTR_INDEX }, ),	# lidt 

	# 0F 00 : LLDT : w LDTR 
	( { 'access':'-w-', 'register':REG_LDTR_INDEX }, ),	# lldt 

	# 0F 01 : LMSW : w CR0 
	( { 'access':'-w-', 'register':REG_CR0_INDEX }, ),	# lmsw 

	# E0, E1, E2 : LOOP : rw ECX 
	# E3 : JCXZ : rw ECX
	( { 'access':'rw-', 'register':REG_ECX_INDEX }, ),# loop 

	# 0F 00 : LTR : w Task Register 
	( { 'access':'-w-', 'register':REG_TR_INDEX }, ),	# ltr 

	# 8F, 58, 1F, 07, 17, 0F A1, 0F A9 : POP : rw ESP 
	# FF, 50, 6A, 68, 0E, 16, 1E, 06, 0F A0, 0F A8 : PUSH : rw ESP 
	( { 'access':'rw-', 'register':REG_ESP_INDEX }, ),	# pop , push

	# 61 : POPAD : rw esp, w edi esi ebp ebx edx ecx eax 
	( { 'access':'rw-', 'register':REG_ESP_INDEX },
	  { 'access':'-w-', 'register':REG_EDI_INDEX },
	  { 'access':'-w-', 'register':REG_ESI_INDEX },
	  { 'access':'-w-', 'register':REG_EBP_INDEX },
	  { 'access':'-w-', 'register':REG_EBX_INDEX },
	  { 'access':'-w-', 'register':REG_EDX_INDEX },
	  { 'access':'-w-', 'register':REG_ECX_INDEX },
	  { 'access':'-w-', 'register':REG_EAX_INDEX } ),	# popad 

	# 9D : POPFD : rw esp, w eflags 
	( { 'access':'rw-', 'register':REG_ESP_INDEX },
	  { 'access':'-w-', 'register':REG_FLAGS_INDEX } ),	# popfd 

	# 60 : PUSHAD : rw esp, r eax ecx edx ebx esp ebp esi edi 
	( { 'access':'rw-', 'register':REG_ESP_INDEX },
	  { 'access':'r--', 'register':REG_EAX_INDEX },
	  { 'access':'r--', 'register':REG_ECX_INDEX },
	  { 'access':'r--', 'register':REG_EDX_INDEX },
	  { 'access':'r--', 'register':REG_EBX_INDEX },
	  { 'access':'r--', 'register':REG_EBP_INDEX },
	  { 'access':'r--', 'register':REG_ESI_INDEX },
	  { 'access':'r--', 'register':REG_EDI_INDEX } ),	# pushad 

	# 9C : PUSHFD : rw esp, r eflags 
	( { 'access':'rw-', 'register':REG_ESP_INDEX },
	  { 'access':'r--', 'register':REG_FLAGS_INDEX } ),	# pushfd 

	# 0F 32 : RDMSR : r ECX, w EDX, w EAX 
	( { 'access':'r--', 'register':REG_ECX_INDEX },
	  { 'access':'-w-', 'register':REG_EDX_INDEX },
	  { 'access':'-w-', 'register':REG_EAX_INDEX } ),	# rdmsr 

	# 0F 33 : RDPMC : r ECX, w EDX, w EAX 
	( { 'access':'r--', 'register':REG_ECX_INDEX },
	  { 'access':'-w-', 'register':REG_EDX_INDEX },
	  { 'access':'-w-', 'register':REG_EAX_INDEX } ),	# rdpmc 

	# 0F 31 : RDTSC : rw EDX, rw EAX 
	( { 'access':'rw-', 'register':REG_EDX_INDEX },
	  { 'access':'rw-', 'register':REG_EAX_INDEX } ),	# rdtsc 

	# F3, F2 ... : REP : rw ECX 
	( { 'access':'rw-', 'register':REG_ECX_INDEX }, ),	# rep 

	# 0F AA : RSM : r CR4, r CR0 
	( { 'access':'r--', 'register':REG_CR4_INDEX }, 
	  { 'access':'r--', 'register':REG_CR0_INDEX } ),	# rsm 

	# 9E : SAHF : r ah, rw eflags (set SF ZF AF PF CF) 
	( { 'access':'r--', 'register':REG_AH_INDEX }, ),	# sahf 

	# TODO: 0F : SGDT : r gdtr 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# sgdt 

	# TODO: 0F : SIDT : r idtr 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# sidt 

	# TODO: 0F : SLDT : r ldtr 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# sldt 

	# TODO: 0F : SMSW : r CR0 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# smsw 

	# TODO: 0F AE : STMXCSR : r MXCSR 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# stmxcsr 

	# TODO: 0F 00 : STR : r TR (task register) 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# str 

	# TODO: 0F 34 : SYSENTER : w cs, w eip, w ss, w esp, r CR0, w eflags
	#         r sysenter_cs_msr, sysenter_esp_msr, sysenter_eip_msr 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# sysenter 

	# TODO: 0F 35 : SYSEXIT : r edx, r ecx, w cs, w eip, w ss, w esp
	#      	   r sysenter_cs_msr 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# sysexit 

	# TODO: 0F 30 : WRMST : r edx, r eax, r ecx 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# wrmsr 

	# TODO: D7 : XLAT : rw al r ebx (ptr) 
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# xlat 
 
	# TODO:  monitor 0f 01 c8 eax r ecx OP_R edx OP_R
	( { 'access':'r--', 'register':REG_EAX_INDEX }, ),	# monitor 

	# TODO: mwait 0f 01 c9 eax r ecx OP_R
	( { 'access':'r--', 'register':REG_EAX_INDEX }, )	# mwait 
)

def implicit_lookup(id):
	''' Implicit operand lookup:
	    Return implicit operand list (tuple of dicts) for ID.
	    ID from the 'implicit' field of the opcode table entry for insn.
	'''
	return implicit_operands[id]
