import libdisasm.isa as ISA

# sizes of various registers
REG_DWORD_SIZE = 4
REG_WORD_SIZE = 2
REG_BYTE_SIZE = 1
REG_MMX_SIZE = 8
REG_SIMD_SIZE = 16
REG_DEBUG_SIZE = 4
REG_CTRL_SIZE = 4
REG_TEST_SIZE = 4
REG_SEG_SIZE = 2
REG_FPU_SIZE = 10
REG_FLAGS_SIZE = 4
REG_FPCTRL_SIZE = 2
REG_FPSTATUS_SIZE = 2
REG_FPTAG_SIZE = 2
REG_EIP_SIZE = 4
REG_IP_SIZE = 2
REG_TASK_SIZE = 80

 
# indexes of significant registers in the register table
# NOTE: these are used in the decoding methods and are not intended
#       to be comprehensive
REG_DWORD_INDEX = 1
REG_EAX_INDEX = 1
REG_ECX_INDEX = 2
REG_EDX_INDEX = 3
REG_EBX_INDEX = 4
REG_ESP_INDEX = 5
REG_EBP_INDEX = 6
REG_ESI_INDEX = 7
REG_EDI_INDEX = 8
REG_WORD_INDEX = 9
REG_AX_INDEX = 9
REG_CX_INDEX = 10
REG_DX_INDEX = 11
REG_BX_INDEX = 12
REG_SP_INDEX = 13
REG_BP_INDEX = 14
REG_SI_INDEX = 15
REG_DI_INDEX = 16
REG_BYTE_INDEX = 17
REG_AL_INDEX = 17
REG_CL_INDEX = 18
REG_DL_INDEX = 19
REG_BL_INDEX = 20
REG_AH_INDEX = 21
REG_CH_INDEX = 22
REG_DH_INDEX = 23
REG_BH_INDEX = 24
REG_MMX_INDEX = 25
REG_SIMD_INDEX = 33
REG_DEBUG_INDEX = 41
REG_CTRL_INDEX = 49
REG_CR0_INDEX = 49
REG_CR4_INDEX = 53
REG_TEST_INDEX = 57
REG_SEG_INDEX = 65
REG_ES_INDEX = 65
REG_CS_INDEX = 66
REG_SS_INDEX = 67
REG_DS_INDEX = 68
REG_FS_INDEX = 69
REG_GS_INDEX = 70
REG_LDTR_INDEX = 71
REG_GDTR_INDEX = 72
REG_FPU_INDEX = 73
REG_ST0_INDEX = 73
REG_ST1_INDEX = 74
REG_ST2_INDEX = 75
REG_ST3_INDEX = 76
REG_ST4_INDEX = 77
REG_ST5_INDEX = 78
REG_ST6_INDEX = 79
REG_ST7_INDEX = 80
REG_FLAGS_INDEX = 81
REG_FPCTRL_INDEX = 82
REG_FPSTATUS_INDEX = 83
REG_FPTAG_INDEX = 84
REG_EIP_INDEX = 85
REG_IP_INDEX = 86
REG_IDTR_INDEX = 87
REG_MXCSG_INDEX = 88
REG_TR_INDEX = 89
REG_CSMSR_INDEX = 90
REG_ESPMSR_INDEX = 91
REG_EIPMSR_INDEX = 92

# NOTE: the MMX register mapping is fixed to the physical registers
# used by the FPU. The floating FP stack does not effect the location
# of the MMX registers, so this aliasing is not 100% accurate.

ia32_registers = (
	None,	# register ID 0 is invalid
	# REG_DWORD_INDEX 
	{ 'size':REG_DWORD_SIZE, 'type':'general, return value', 'alias_of':None, 'name':'eax' },
	{ 'size':REG_DWORD_SIZE, 'type':'general, rep counter', 'alias_of':None, 'name':'ecx' },
	{ 'size':REG_DWORD_SIZE, 'type':'general', 'alias_of':None, 'name':'edx' },
	{ 'size':REG_DWORD_SIZE, 'type':'general', 'alias_of':None, 'name':'ebx' },
	# REG_ESP_INDEX 
	{ 'size':REG_DWORD_SIZE, 'type':'general, stack pointer', 'alias_of':None, 'name':'esp' },
	{ 'size':REG_DWORD_SIZE, 'type':'general, frame pointer', 'alias_of':None, 'name':'ebp' },
	{ 'size':REG_DWORD_SIZE, 'type':'general, rep source', 'alias_of':None, 'name':'esi' },
	{ 'size':REG_DWORD_SIZE, 'type':'general, rep dest', 'alias_of':None, 'name':'edi' },
	# REG_WORD_INDEX 
	{ 'size':REG_WORD_SIZE, 'type':'general, return value', 'alias_of':REG_EAX_INDEX, 'alias_shift':0, 'name':'ax' },
	{ 'size':REG_WORD_SIZE, 'type':'general, rep counter', 'alias_of':REG_ECX_INDEX, 'alias_shift':0, 'name':'cx' },
	{ 'size':REG_WORD_SIZE, 'type':'general', 'alias_of':REG_EDX_INDEX, 'alias_shift':0, 'name':'dx' },
	{ 'size':REG_WORD_SIZE, 'type':'general', 'alias_of':REG_EBX_INDEX, 'alias_shift':0, 'name':'bx' },
	{ 'size':REG_WORD_SIZE, 'type':'general, stack pointer', 'alias_of':REG_ESP_INDEX, 'alias_shift':0, 'name':'sp' },
	{ 'size':REG_WORD_SIZE, 'type':'general, frame pointer', 'alias_of':REG_EBP_INDEX, 'alias_shift':0, 'name':'bp' },
	{ 'size':REG_WORD_SIZE, 'type':'general, rep source', 'alias_of':REG_ESI_INDEX, 'alias_shift':0, 'name':'si' },
	{ 'size':REG_WORD_SIZE, 'type':'general, rep dest', 'alias_of':REG_EDI_INDEX, 'alias_shift':0, 'name':'di' },
	# REG_BYTE_INDEX 
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_EAX_INDEX, 'alias_shift':0, 'name':'al' },
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_ECX_INDEX, 'alias_shift':0, 'name':'cl' },
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_EDX_INDEX, 'alias_shift':0, 'name':'dl' },
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_EBX_INDEX, 'alias_shift':0, 'name':'bl' },
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_EAX_INDEX, 'alias_shift':8, 'name':'ah' },
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_ECX_INDEX, 'alias_shift':8, 'name':'ch' },
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_EDX_INDEX, 'alias_shift':8, 'name':'dh' },
	{ 'size':REG_BYTE_SIZE, 'type':'general', 'alias_of':REG_EBX_INDEX, 'alias_shift':8, 'name':'bh' },
	# REG_MMX_INDEX 
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST0_INDEX, 'alias_shift':0, 'name':'mm0' },
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST1_INDEX, 'alias_shift':0, 'name':'mm1' },
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST2_INDEX, 'alias_shift':0, 'name':'mm2' },
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST3_INDEX, 'alias_shift':0, 'name':'mm3' },
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST4_INDEX, 'alias_shift':0, 'name':'mm4' },
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST5_INDEX, 'alias_shift':0, 'name':'mm5' },
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST6_INDEX, 'alias_shift':0, 'name':'mm6' },
	{ 'size':REG_MMX_SIZE, 'type':'simd', 'alias_of':REG_ST7_INDEX, 'alias_shift':0, 'name':'mm7' },
	# REG_SIMD_INDEX 
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm0' },
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm1' },
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm2' },
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm3' },
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm4' },
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm5' },
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm6' },
	{ 'size':REG_SIMD_SIZE, 'type':'simd', 'alias_of':None, 'name':'xmm7' },
	# REG_DEBUG_INDEX 
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr0'},
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr1'},
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr2'},
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr3'},
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr4'},
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr5'},
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr6'},
	{ 'size':REG_DEBUG_SIZE, 'type':'system', 'alias_of':None,'name':'dr7'},
	# REG_CTRL_INDEX 
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr0'},
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr1'},
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr2'},
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr3'},
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr4'},
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr5'},
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr6'},
	{ 'size':REG_CTRL_SIZE, 'type':'system', 'alias_of':None, 'name':'cr7'},
	# REG_TEST_INDEX 
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr0'},
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr1'},
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr2'},
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr3'},
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr4'},
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr5'},
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr6'},
	{ 'size':REG_TEST_SIZE, 'type':'system', 'alias_of':None, 'name':'tr7'},
	# REG_SEG_INDEX 
	{ 'size':REG_SEG_SIZE, 'type':'segment', 'alias_of':None, 'name':'es' },
	{ 'size':REG_SEG_SIZE, 'type':'segment', 'alias_of':None, 'name':'cs' },
	{ 'size':REG_SEG_SIZE, 'type':'segment', 'alias_of':None, 'name':'ss' },
	{ 'size':REG_SEG_SIZE, 'type':'segment', 'alias_of':None, 'name':'ds' },
	{ 'size':REG_SEG_SIZE, 'type':'segment', 'alias_of':None, 'name':'fs' },
	{ 'size':REG_SEG_SIZE, 'type':'segment', 'alias_of':None, 'name':'gs' },
	# REG_LDTR_INDEX 
	{ 'size':REG_DWORD_SIZE, 'type':'system', 'alias_of':None, 'name':'ldtr' },
	# REG_GDTR_INDEX 
	{ 'size':REG_DWORD_SIZE, 'type':'system', 'alias_of':None, 'name':'gdtr' },
	# REG_FPU_INDEX 
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(0)' },
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(1)' },
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(2)' },
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(3)' },
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(4)' },
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(5)' },
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(6)' },
	{ 'size':REG_FPU_SIZE, 'type':'floating point', 'alias_of':None, 'name':'st(7)' },
	# REG_FLAGS_INDEX : 81 
	{ 'size':REG_FLAGS_SIZE, 'type':'condition codes', 'alias_of':None, 'name':'eflags' }, 
	# REG_FPCTRL_INDEX  : 82
	{ 'size':REG_FPCTRL_SIZE, 'type':'floating point, system', 'alias_of':None, 'name':'fpctrl' }, 
	# REG_FPSTATUS_INDEX : 83
	{ 'size':REG_FPSTATUS_SIZE, 'type':'floating point, system', 'alias_of':None, 'name':'fpstat' },
	# REG_FPTAG_INDEX : 84 
	{ 'size':REG_FPTAG_SIZE, 'type':'floating point, system', 'alias_of':None, 'name':'fptag' }, 
	# REG_EIP_INDEX : 85 
	{ 'size':REG_EIP_SIZE, 'type':'program counter', 'alias_of':None, 'name':'eip' },
	# REG_IP_INDEX : 86 
	{ 'size':REG_IP_SIZE, 'type':'program counter', 'alias_of':REG_EIP_INDEX, 'alias_shift':0, 'name':'ip' },
	# REG_IDTR_INDEX : 87 
	{ 'size':REG_DWORD_SIZE, 'type':'system', 'alias_of':None, 'name':'idtr' },
	# REG_MXCSG_INDEX : SSE Control Reg : 88 
	{ 'size':REG_DWORD_SIZE, 'type':'system, simd', 'alias_of':None, 'name':'mxcsr' },
	# REG_TR_INDEX : Task Register : 89 
	{ 'size':REG_TASK_SIZE, 'type':'system', 'alias_of':None, 'name':'tr' },
	# REG_CSMSR_INDEX : SYSENTER_CS_MSR : 90 
	{ 'size':REG_DWORD_SIZE, 'type':'system', 'alias_of':None, 'name':'cs_msr' },
	# REG_ESPMSR_INDEX : SYSENTER_ESP_MSR : 91 
	{ 'size':REG_DWORD_SIZE, 'type':'system', 'alias_of':None, 'name':'esp_msr' },
	# REG_EIPMSR_INDEX : SYSENTER_EIP_MSR : 92 
	{ 'size':REG_DWORD_SIZE, 'type':'system', 'alias_of':None, 'name':'eip_msr' }
)

def register_lookup(id):
	''' Register lookup:
	    Return register decriptor (dictionary) for ID.
	    ID is an operand value from opcode tables or modr/m decoding.
	'''
	return ia32_registers[id]

# runtime cache of register objects
ia32_register_objects = [ None ] * 92

def register_factory(id):
	if ia32_register_objects[id]:
		return ia32_register_objects[id]

	descr = ia32_registers[id]
	alias = None
	if descr['alias_of']:
		alias = register_factory(descr['alias_of'])
	reg = ISA.Register(descr, alias, id)
	ia32_register_objects[id] = reg

	return reg
